use axum::{
    body::Bytes,
    extract::{ContentLengthLimit, Extension, Path},
    handler::Handler,
    http::StatusCode,
    routing::{get},
    Router,
};
use std::{
    collections::HashMap,
    sync::{Arc, RwLock},
};
use tower_http::{
   compression::CompressionLayer,// trace::TraceLayer,
};

pub type SharedState = Arc<RwLock<State>>;

#[derive(Default)]
pub struct State {
  pub db: HashMap<String, Bytes>,
}

async fn kv_get(
    Path(key): Path<String>,
    Extension(state): Extension<SharedState>,
) -> Result<Bytes, StatusCode> {
    let db = &state.read().unwrap().db;

    if let Some(value) = db.get(&key) {
        Ok(value.clone())
    } else {
        Err(StatusCode::NOT_FOUND)
    }
}

async fn kv_set(
    Path(key): Path<String>,
    ContentLengthLimit(bytes): ContentLengthLimit<Bytes, { 1024 * 5_000 }>, // ~5mb
    Extension(state): Extension<SharedState>,
) {
    state.write().unwrap().db.insert(key, bytes);
}

async fn list_keys(Extension(state): Extension<SharedState>) -> String {
    let db = &state.read().unwrap().db;

    db.keys()
        .map(|key| key.to_string())
        .collect::<Vec<String>>()
        .join("\n")
}

pub fn router_handlers(web_router: Router) -> Router {
	let mut webrouter = web_router.to_owned();
	webrouter = crate::handlers::kratos::def_handlers(webrouter);
  webrouter
    .route(
				"/:key",
				// Add compression to `kv_get`
				get(kv_get.layer(CompressionLayer::new()))
						// But don't compress `kv_set`
						.post(kv_set),
		)
		.route("/keys", get(list_keys))
}